// Copyright (c) 2014-2019 Sombusta, All Rights Reserved.
// MIT LICENSE // Epic VR Template Convert C++ Open Source Project.

#include "TP_VirtualRealityPawn_Motion.h"
#include "TP_MotionController.h"
#include "Engine/World.h"
#include "TimerManager.h"
#include "Engine.h"
#include "Macros/Macros.h"
#include "Actors/MovingObject.h"
#include "Kismet/GameplayStatics.h"
#include "Kismet/KismetMathLibrary.h"
#include "Kismet/KismetSystemLibrary.h"
#include "HeadMountedDisplayFunctionLibrary.h"
#include "GameFramework/InputSettings.h"
#include "Camera/CameraComponent.h"
#include "Components/InputComponent.h"
#include "Components/SceneComponent.h"
#include "MotionControllerComponent.h"

// Sets default values
ATP_VirtualRealityPawn_Motion::ATP_VirtualRealityPawn_Motion()
{
	PrimaryActorTick.bCanEverTick = true;

	// Create Components Initialize
	RootScene = CreateDefaultSubobject<USceneComponent>(TEXT("DefaultSceneRoot"));
	CameraBase = CreateDefaultSubobject<USceneComponent>(TEXT("VROrigin"));
	VRCamera = CreateDefaultSubobject<UCameraComponent>(TEXT("Camera"));
	PlayerCollision = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("Cylinder"));

	RootComponent = RootScene;
	CameraBase->SetupAttachment(RootComponent);
	VRCamera->SetupAttachment(CameraBase);
	PlayerCollision->SetupAttachment(CameraBase);
}

// Called when the game starts or when spawned
void ATP_VirtualRealityPawn_Motion::BeginPlay()
{
	Super::BeginPlay();

	bBlockInput = false;
	// // Setup Player Height for various Platforms (PS4, Vive, Oculus)
	FName DeviceName = UHeadMountedDisplayFunctionLibrary::GetHMDDeviceName();

	if (DeviceName == "SteamVR" || DeviceName == "OculusHMD")
	{
		// Windows (Oculus / Vive)
		UHeadMountedDisplayFunctionLibrary::SetTrackingOrigin(EHMDTrackingOrigin::Floor);
	}
	else
	{
		// PS4
		UHeadMountedDisplayFunctionLibrary::SetTrackingOrigin(EHMDTrackingOrigin::Eye);
		CameraBase->AddLocalOffset(FVector(0.0f, 0.0f, DefaultPlayerHeight));	
	}
	

	// Spawn and attach both motion controllers

	const FTransform SpawnTransform = FTransform(FRotator(0.0f, 0.0f, 0.0f), FVector(0.0f, 0.0f, 0.0f), FVector(1.0f, 1.0f, 1.0f)); // = FTransform::Identity;
	FAttachmentTransformRules AttachRules(EAttachmentRule::SnapToTarget, EAttachmentRule::SnapToTarget, EAttachmentRule::KeepWorld, false);

	// "Hand" is available by checking "Expose on Spawn" in the variable on BP_MotionController.
	// Expose on Spawn Variable parameter setup in c++ -> Use SpawnActorDeferred
	LeftController = GetWorld()->SpawnActorDeferred<ATP_MotionController>(ATP_MotionController::StaticClass(), SpawnTransform, this, nullptr, ESpawnActorCollisionHandlingMethod::AlwaysSpawn);
	if (LeftController)
	{
		LeftController->Hand = EControllerHand::Left;
		LeftController->SetWidgetInteractionBlueprint(WidgetInteractionBlueprint);// Tell it where the widget interaction blueprint is
		LeftController->FinishSpawning(SpawnTransform); // Call begin play and contruction scripts
		LeftController->AttachToComponent(CameraBase, AttachRules);
	}

	RightController = GetWorld()->SpawnActorDeferred<ATP_MotionController>(ATP_MotionController::StaticClass(), SpawnTransform, this, nullptr, ESpawnActorCollisionHandlingMethod::AlwaysSpawn);
	if (RightController)
	{
		RightController->Hand = EControllerHand::Right;
		RightController->SetWidgetInteractionBlueprint(WidgetInteractionBlueprint);// Tell it where the widget interaction blueprint is
		RightController->FinishSpawning(SpawnTransform);// Call begin play and contruction scripts
		RightController->AttachToComponent(CameraBase, AttachRules);
	}
	LeftController->SetOtherController(RightController->GetMotionControllerMesh());
	RightController->SetOtherController(LeftController->GetMotionControllerMesh()); 
}

// Called every frame
void ATP_VirtualRealityPawn_Motion::Tick(float DeltaTime)
{
	
	Super::Tick(DeltaTime);

	FVector VRLocation = VRCamera->GetComponentLocation();
	if (!bHasChangedHeight&&VRLocation.Z>30)
	{
		bHasChangedHeight = true;
		ChangeSceneHeight();
		
	}
	
	PlayerCollision->SetWorldLocation(FVector(VRLocation.X, VRLocation.Y, 15));
	//Check if stepping on the conveyer belt.
	if (VRLocation.Y < MinimumY) {
		bIsInGoodPosition = false;
	}
	else {
		bIsInGoodPosition = true;
		
	}
}

// Called to bind functionality to input
void ATP_VirtualRealityPawn_Motion::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
	Super::SetupPlayerInputComponent(PlayerInputComponent);

	// Set up gameplay key bindings
	check(PlayerInputComponent);

	// Bind Recenter VR events
	PlayerInputComponent->BindAction("ResetVR", IE_Pressed, this, &ATP_VirtualRealityPawn_Motion::StartTimerOfResettingVR);
	PlayerInputComponent->BindAction("ResetVR", IE_Released, this, &ATP_VirtualRealityPawn_Motion::StopTimerOfResettingVR);

	// Bind Grab events
	PlayerInputComponent->BindAction("GrabLeft", IE_Pressed, this, &ATP_VirtualRealityPawn_Motion::GrabActor_Left);
	PlayerInputComponent->BindAction("GrabLeft", IE_Released, this, &ATP_VirtualRealityPawn_Motion::ReleaseActor_Left);
	PlayerInputComponent->BindAction("GrabRight", IE_Pressed, this, &ATP_VirtualRealityPawn_Motion::GrabActor_Right);
	PlayerInputComponent->BindAction("GrabRight", IE_Released, this, &ATP_VirtualRealityPawn_Motion::ReleaseActor_Right);

	// Bind Input Axises
	PlayerInputComponent->BindAxis(TEXT("MotionControllerThumbLeft_Y"));
	PlayerInputComponent->BindAxis(TEXT("MotionControllerThumbLeft_X"));
	PlayerInputComponent->BindAxis(TEXT("MotionControllerThumbRight_Y"));
	PlayerInputComponent->BindAxis(TEXT("MotionControllerThumbRight_X"));
}

void ATP_VirtualRealityPawn_Motion::ResetingVR()
{
	HasResetVR = true;
	bool lHasGrabbedLeft = HasGrabbedLeft;
	bool lHasGrabbedRight = HasGrabbedRight;
	UHeadMountedDisplayFunctionLibrary::ResetOrientationAndPosition();
	bHasChangedHeight = false;
	if (lHasGrabbedLeft) {
		GrabActor_Left();
	}
	if (lHasGrabbedRight) {
		GrabActor_Right();
	}
}

void ATP_VirtualRealityPawn_Motion::OffsetComponentByChangedHeight(USceneComponent* Component)
{
	if (!ensure(Component)) { return; }
	Component->AddWorldOffset(FVector(0, 0, SceneMovement));
	//Changing height of Component by SceneMovement
	MovableActors.AddUnique(Component->GetOwner());
}

void ATP_VirtualRealityPawn_Motion::GrabActor_Left()
{
	HasGrabbedLeft = true;
	LeftController->GrabActor();
}

void ATP_VirtualRealityPawn_Motion::ReleaseActor_Left()
{
	HasGrabbedLeft = false;
	LeftController->ReleaseActor();
}

void ATP_VirtualRealityPawn_Motion::GrabActor_Right()
{
	HasGrabbedRight = true;
	RightController->GrabActor();
}

void ATP_VirtualRealityPawn_Motion::ReleaseActor_Right()
{
	HasGrabbedRight = false;
	RightController->ReleaseActor();
}



bool ATP_VirtualRealityPawn_Motion::IsInGoodPosition()
{
	return bIsInGoodPosition;
}

void ATP_VirtualRealityPawn_Motion::StartTimerOfResettingVR()
{
	GetWorldTimerManager().SetTimer(ResetVRTimerHandle, this, &ATP_VirtualRealityPawn_Motion::ResetingVR , MinimumTimeForHoldToResetVR,false);
}

void ATP_VirtualRealityPawn_Motion::StopTimerOfResettingVR()
{
	GetWorldTimerManager().ClearTimer(ResetVRTimerHandle);
}

bool ATP_VirtualRealityPawn_Motion::WorldType__InGameInstanceWorld(UObject* WorldContextObject)
{
	if (!WorldContextObject) return false;

	//using a context object to get the world!
	UWorld* const World = GEngine->GetWorldFromContextObject(WorldContextObject, EGetWorldErrorMode::LogAndReturnNull);
	if (!World) return false;

	return (World->WorldType == EWorldType::Game);
}

void ATP_VirtualRealityPawn_Motion::ChangeSceneHeight()
{
	//Gets and logs the camera height
	FVector VRLocation = VRCamera->GetComponentLocation();
	//Height is VRLocation.Z
	if (!(MovableActors.IsValidIndex(0)) || !MovableActors[0]) { return; }
	

	//Ofsets all the moving objects.
	TArray<AActor*> MovingObjects;
	UGameplayStatics::GetAllActorsOfClass(GetWorld(), AMovingObject::StaticClass(), MovingObjects);
	for (AActor* CurrentActor : MovingObjects) {
		AMovingObject* CurrentMovingObject = Cast<AMovingObject>(CurrentActor);
		if (CurrentMovingObject->GetRootComponent()->IsSimulatingPhysics()) {
			CurrentMovingObject->AddActorWorldOffset(FVector(0, 0, (-SceneMovement + (VRLocation.Z - DefaultPlayerHeight))+2));
		}
	}
	//Offsets the conveyers and bins.
	if (MovableActors.Num()==0) { return; }
	for (AActor* CurrentActor : MovableActors) {
		if (CurrentActor->IsValidLowLevel()) {
			CurrentActor->AddActorWorldOffset(FVector(0, 0, -SceneMovement + (VRLocation.Z - DefaultPlayerHeight) - 4));
		}
	}

	//updates how much the objects are moved from their starting height.
	SceneMovement = (VRLocation.Z - DefaultPlayerHeight);
}

ATP_MotionController* ATP_VirtualRealityPawn_Motion::GetRightController()
{
	return RightController;
}

ATP_MotionController* ATP_VirtualRealityPawn_Motion::GetLeftController()
{
	return LeftController;
}